"""
Functions and class definitions for handling experimental data.
Author: Natalia Spitha, 6/3/2019
"""

import numpy as np
from scipy import signal

from KinetiKit import sim
from KinetiKit.units import units, ns
from KinetiKit import kit as kin_kit
import warnings
import numpy
import copy

__all__ = ['data_from_SPCM']


class data_from_SPCM(object):
    """
    Data object created by importing one time-resolved .asc signal data file from
    a Becker & Hickl SPC-130 single photon counting module (SPCM software)
    
    Can be used to represent a generic TRPL text file import, but the 
    `metadata` keyword cannot be used meaningfully in this case.
    
    Parameters
    ----------
    filename : string
        the file path of interest
    
    Optional Parameters
    -------------------
    key : string
        an identifier of choice for the object, which can be later used in 
        labels etc. Default is `None`.
    time_unit : string from ``KinetiKit.units.units`` dictionary keys
        The unit in which time is expressed in the data file. Default is`'ns'`.
    coll : float
        Collection time of the experiment in seconds. If `metadata` is `True`, 
        and the specific Becker & Hickl SPCM format is used,`coll` is 
        over-ridden by the collection time embedded in the file. Default is 1.
    weigh_by_coll : boolean
        If `True`, the signal values (y) of the object are automatically 
        divided by the collection time to yield counts-per-second units. 
        Default is `True`.
    metadata : boolean
        If `True` the collection time is obtained from inside the .asc file.
        For this data object,`metadata=True` only works for Becker & Hickl SPCM
        data files. Default is  `False`.
    pulse_power : float
        pulse power information (in any desired units) for the data. Default is
        `None`.
    cw_power : float
        continuous-wave power information (in any desired units) for the data.
        Default is `None`.
    wavelength : float
        wavelength information (in any desired units) for the data. Default is
        `None`.
    delimiter : string
        Delimiter in the text file. Default is `','`.
    unpack : boolean
        See `unpack` argument of ``numpy.genfromtxt``.
    skip_h : integer
        See `skip_header` argument of ``numpy.genfromtxt``.
    skip_f : integer
        See `skip_footer` argument of ``numpy.genfromtxt``.
        
    """
    
    def __init__(self, filename, key=None, time_unit='ns', coll=1,
                 metadata=False, weigh_by_coll=True, 
                 pulse_power=None, cw_power=None, wavelength=None,
                 delimiter=',', unpack=True, skip_h=10, skip_f=1):
        
        self.filename = filename
        self.time_unit = time_unit
        self.key = key # identifier
        self.weigh_by_coll = weigh_by_coll
        self.skip_h = skip_h; self.skip_f = skip_f
        
        # initializing some attributes
        self.dark_subtracted=False
        self.weighed_by_coll = False
        
        # If aquisition information was included in file, this part extracts
        # the aquisition time:
        if metadata:
            with open(self.filename) as file:
                line_col = file.readlines()[45]
                self.coll = float(line_col.split(',')[2].split(']')[0])
    
        else:
            self.coll = coll
                
        self.x, self.y = np.genfromtxt(filename, delimiter=',', unpack=unpack, skip_header=skip_h, skip_footer=skip_f) 
        
        if weigh_by_coll:
            self.y /= self.coll
            self.weighed_by_coll =True
                
        self.x -= self.x[0]
        self.x *= units[time_unit]
        
        self.pulse_power = pulse_power
        self.cw_power = cw_power
        
        self.wavelength = wavelength
        
    
    def dark_subtract(self, dark_counts, method='average'):
        """
        Accepts a float, array, or data object representing dark counts, and
        performs a dark count subtraction on the current data object.
        
        This function does not check for intensity or time units; ensure that
        the count units (recommended: Counts Per Second) are consistent between
        the data object and the dark count instance.
        
        Parameters
        -------------------
        dark_counts : float, array, or KinetiKit.data.lib.data object
            If dark_counts is an array, it should have the same length as 
            the y array of the data object.
        method : 'average' or 'elementwise', optional
        
        """
        
        if not self.dark_subtracted:
            
            if isinstance(dark_counts, numpy.float) or isinstance(dark_counts, numpy.int):
                self.y -= dark_counts
            elif isinstance(dark_counts, numpy.ndarray):
                if method == 'elementwise': 
                    self.y -= dark_counts
                elif method == 'average':
                    self.y -= np.average(dark_counts)
            elif isinstance(dark_counts, data_from_SPCM):
                if self.weighed_by_coll==False or dark_counts.weighed_by_coll==False:
                    warnings.warn('Dark subtraction was performed. However, we recommend that you convert Counts to Counts Per Second by weighing data objects by the collection time.', UserWarning)
                if method == 'elementwise':
                    self.y -= dark_counts.y
                elif method == 'average':
                    self.y -= np.average(dark_counts.y)
            else:
                return
            self.dark_subtracted=True
            
        else: print('Dark subtraction has already been performed')
        
        
    def smooth(self, window_length=17, polyorder=2, mode='nearest'):
        """
        Wrapper function for the ``scipy.signal.savgol_filter`` smoothing 
        function with limited relevant arguments
        
        """
                                
        self.y = signal.savgol_filter(self.y, window_length = window_length, 
                                    polyorder = polyorder, mode = mode)
    
    def interp(self, t):
        """
        Interpolates the y-values of the imported object along a time axis.
        Interpolation is most accurate if the first and last element of the 
        time array match the first and last element of the object's x-data.
        
        Parameters
        ----------
        t : 1-D array or dictionary of the type ``sim.time.linear()``
            Defines the time axis along which to interpolate.

        """
        if isinstance(t, numpy.ndarray):
            time_array = t
            self.y = np.interp(time_array, self.x, self.y)
            self.x = time_array
            
        if isinstance(t, dict):
            time_array = t['array'][::t['subsample']]
            self.y = np.interp(time_array, self.x, self.y)
            self.x = time_array
        
    
    def remove_zeros(self):
        """
        Replaces any negative or zero elements of the object's y-data with the 
        minimum nonzero positive value of the y-data array.
        """
        
        y=self.y
        ymin = min(y[np.where(y>0)])
        ynew = y.copy()
        ynew[np.where(ynew<=0)] = ymin
        self.y = ynew
    
    def extract_average(self, t_start=12*ns, t_end=12.5*ns):
        """
        Returns the average y-value of the object defined between two x-values.
        The x-values should be given in seconds and `t_start` must be less than
        `t_end`.
        
        """
        
        i_start, t_start = kin_kit.find_nearest(self.x, t_start)
        i_end, t_end = kin_kit.find_nearest(self.x, t_end)
        return np.average(self.y[i_start:i_end])
        
    def extract_peak(self, avgnum=5):
        """
        Returns the peak y-value of the object, averaging between a defined 
        number of points (`avgnum`) to control for noise.
        
        """
        peak_values = []
        y = list(self.y)
        for i in range(avgnum):
            peak_values.append(max(y))
            y.remove(max(y))
        return np.average(peak_values)
    
    def copy(self):
        """
        Returns an independent copy of the data object and all its attributes
        at the time this function is called.
        """
        return copy.deepcopy(self)
        
        



"""
Functions specializing on array shifting and alignment
"""

import sys
import os
import csv

import numpy as np
import matplotlib.pyplot as plt

def find_nearest(array, value):
    """
    Returns element and index of array that is nearest to value.
    
    Parameters
    ----------
    array : arr
        The array from which a value and index will be returned
    value : float
        The value that the returned element of `arr` will be nearest to
        
    Returns
    -------
    idx : int
        The index of `arr` corresponding to the nearest value to `value`
    nearest_val : float
        The nearest value of the array to `value`
    """
    
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    nearest_val = array[idx]
    return idx, nearest_val

def roll_by_array_shift(arrays, refarray, shift, direction='forward'):
    """
    Performs ``numpy.roll(arrays)`` by the same number of points that 
    correspond to a `shift` in a reference array (`refarray`).
    
    Assumes linear step size for `refarray` and `arrays`.
    
    Parameters
    ----------
    arrays : 1D or 2D array
        Arrays whose values will be shifted.
    refarray : 1D array
        Array to whose values the `shift` pertains. Must have a 1:1 element
        correspondence with rollarray, even though it can be smaller. 
        `rollarray` will be rolled by the same number of elements that add up 
        to `shift` in `refarray`
    shift : float
        Desired shift in `refarray` values.
    direction : str, 'forward' or 'back'
        Determines the direction of rolling. Default is 'forward.'
    
    Returns
    -------
    rolled_array : arr
        Shifted array
    """
    rolled_arrays = arrays.copy()
    
    offset = shift + refarray[0]
    idx, val = find_nearest(refarray, offset)
    if direction == 'forward':
        rolled_arrays = np.roll(rolled_arrays, idx, axis=-1)
    elif direction == 'back':
        rolled_arrays = np.roll(rolled_arrays, -idx, axis=-1)    
    else:
        raise ValueError('Direction must be \'forward\' or \'back\'.')
        #issue_error
    
    return rolled_arrays

 
def align_by_max(arrays, refarray_x, refarray_y=None, avgnum=1, value=0):
    """
    Determines how many indices refarray_y must be rolled by so that its max.
    region aligns with `value` in `refarray_x', and then rolls each array in 
    `arrays` by the same amount. The number of points considered to define the
    maximum "region" of refarray_y is defined by `avgnum`.
    
    Parameters
    ----------
    arrays : 1-D or 2-D array
    refarray_x : 1-D array
    
    Optional Parameters
    -------------------
    refarray_y : 1-D array
        Default is None, in which case refarray_y is defined as the first array
        in  `arrays`.
    avgnum : integer
        The integer `avgnum` is used as follows: The top `avgnum` elements of 
        each array_list are considered, which each correspond to a unique value
        in `refarray`. For each array to be rolled, the average of the 
        `refarray` values corresponding to the `avgnum` maxima is considered, 
        and ultimately determines the horizontal shift. Default is 1. Setting 
        `avgnum` > 1 can be useful if the y array is noisy and a single maximum
        element is not representative of its true maximum.
    value : float
        Default is zero. If the exact value does not exist in `refarray_x`, 
        then the nearest value on the array will be chosen.
    
    Returns
    -------
    rolled_arrays : list of arrays or 2D array
        Set of aligned arrays.
    """
    rolled_arrays = arrays.copy()
    
    if refarray_y is None:
        if rolled_arrays.ndim == 2:
            refarray_y = rolled_arrays[0].copy()
        elif rolled_arrays.ndim == 1:
            refarray_y = rolled_arrays.copy()
        else: 
            print('Function only accepts 1D and 2D arrays.')
            #issue_error
            return
    
    sacrificial = refarray_y.copy()
    refvalues_of_maxes = np.array([])
        
    for i in range(0,avgnum):
        current_max = sacrificial.max()
        max_idx = np.where(refarray_y==current_max)[0][0]
        current_refvalue = refarray_x[max_idx]
        refvalues_of_maxes = np.append(refvalues_of_maxes, [current_refvalue])
        sacrificial[np.where(sacrificial==current_max)[0][0]] = -sys.maxsize
        
    avg_ref_value = np.average(refvalues_of_maxes)
    max_idx = find_nearest(refarray_x, avg_ref_value)[0]
        
    val_idx, val = find_nearest(refarray_x, value)
        
    shift_idx = int(max_idx - val_idx)
    
    rolled_arrays = np.roll(rolled_arrays, -shift_idx, axis=-1)
    
    return rolled_arrays


def align_by_steep(arrays, refarray_x, refarray_y=None, avgnum=3, value=0):
    """
    Analogous function to  ``align_by_max``: Rolls each array in `arrays` by 
    the index required to align the steepest region of `refarray_y` with a 
    `value` in `refarray_x`. The number of points considered to define the
    steepest "region" of refarray_y is defined by `avgnum`.
    
    Required Parameters
    -------------------
    arrays : array or list of 1D arrays
    refarray_x : 1-D array
    
    Optional Parameters
    -------------------
    refarray_y : 1-D array
        Default is None, in which case refarray_y is defined as the first array
        in  `arrays`.
    value : float
        Default is zero. If the exact value does not exist in `refarray_x`, 
        then the nearest value on the array will be chosen.
    
    Returns
    -------
    rolled_arrays : list of arrays or 2D array
        Set of aligned arrays.
    """
     
    rolled_arrays = arrays.copy()
    
    if refarray_y is None:
        if rolled_arrays.ndim == 2:
            refarray_y = rolled_arrays[0].copy()
        elif rolled_arrays.ndim == 1:
            refarray_y = rolled_arrays.copy()
        else: 
            print('Function only accepts 1D and 2D arrays.')
            #issue_error
            return
        
    difs = np.zeros(len(refarray_y))
    for i in range(len(refarray_y)-1):
        difs[i] = refarray_y[i]-refarray_y[i-2*avgnum]
    difs = np.roll(difs, -avgnum)
    
    steep_idx = np.argmax(difs)
    val_idx, val = find_nearest(refarray_x, value)
    shift_idx = int(steep_idx - val_idx)
    
    rolled_arrays = np.roll(rolled_arrays, -shift_idx, axis=-1)
        
    return rolled_arrays


import numpy as np
from scipy.signal import convolve

from KinetiKit.units import units, MHz, fs, nm, uW, ps
import KinetiKit.kit as kin_kit

__all__ = ['convolve_irf']

# --- IRF Functions
def convolve_irf(pl, t, fwhm=40 * ps):
    irf = build_irf(t, fwhm=fwhm)
    if pl.ndim == 2:
        irf = irf[np.newaxis]
    return convolve_cyclic_boundary(pl, irf)


def build_irf(t, fwhm=40 * ps):
    tirf = t[t <= 3 * fwhm]
    tirf -= tirf.mean()
    irf = kin_kit.Gauss(tirf, 1, 0, fwhm)
    return irf / irf.sum()


def convolve_cyclic_boundary(data, irf):
    m = irf.size
    out = np.concatenate(
        (data[..., -m:], data[..., :], data[..., :m]), axis=-1)
    out = convolve(out, irf, mode='same')
    return out[..., m:-m]



"""
Fitting library 

Library of functions and classes that relate and compare data with simulations
for the purposes of fitting.

"""
import numpy as np
import scipy as sp

from KinetiKit import sim
from KinetiKit.units import units, ps
from KinetiKit.settings import settings, counter
import KinetiKit.kit as kin_kit


def elementwise_diff(data_arrays, sim_arrays, norm=True, comparison='linear',
                     absolute=False,):
    """
    Returns a list of arrays of the relative differences between pairs of 
    experimental data and simulated data.
    
    Parameters
    ----------
    data_arrays : array or list of 1D arrays
        Experimental data.
    sim_arrays : array or list of 1D arrays
        Simulated data. Must have same shape as `data_arrays`. Comparison will be
        made item by item.
    norm : boolean, optional
        Determines whether the comparison between data and simulation will be
        between normalized or un-normalized arrays. Default is True.
    absolute : boolean, optional
        Determines whether the function will return absolute differences (True)
        or the differences divided by the values of `data_arrays` (False). 
        Default is False.
    comparison : string
        Must be 'linear' or 'log'. Determines whether the function returns the
        difference in logs or the absolute difference between the two sets
        of arrays.
        
    Returns
    -------
    diffs : array or list of 1D arrays
        Returns the element-by-element relative differences between `data_arrays`
        and `sim_arrays` in the form of an array of the same shape as the input
        arrays. 
    """
    if np.all(sim_arrays==0):
        diffs = np.ones(sim_arrays.shape)*100000
        return diffs
    
    #print(len(data_arrays), len(sim_arrays))
    
    alert=False
    if norm:
        for i, array in enumerate(data_arrays):
            data_arrays = kin_kit.normalized(data_arrays, alert)
            sim_arrays = kin_kit.normalized(sim_arrays, alert)

    if absolute==False:
        diffs = abs( data_arrays[np.where(data_arrays !=0)] 
        - sim_arrays[np.where(data_arrays !=0)] ) / data_arrays[np.where(data_arrays !=0)]
        
    else:
        if comparison == 'linear':
            diffs = abs( data_arrays[np.where(data_arrays !=0)] - sim_arrays[np.where(data_arrays !=0)] ) 
        elif comparison == 'log':
            diffs = abs(np.log10(data_arrays[np.where(data_arrays > 0)]) 
            - np.log10(sim_arrays[np.where(data_arrays > 0)]))
        else: 
            print('Comparison must be linear or log.')
            return
    #print('lendiffs', len(diffs))
    return diffs

def slice_by_time(arrays, timearray, lowlim=None, hilim=None):
    """
    Slices a set of arrays along values of a reference array with the same
    'long' dimension.
    """
    
    sliced_arrays = arrays.copy()
    
    lowidx, lowval = kin_kit.find_nearest(timearray, lowlim)
    hiidx, hival = kin_kit.find_nearest(timearray, hilim)
    
    if lowidx > hiidx:
        sliced_arrays = np.roll(sliced_arrays, -lowidx, axis=-1)
        hiidx += len(timearray) - lowidx
        if len(sliced_arrays.shape) != 1:
            return sliced_arrays[:, 0:hiidx+1]
        else:
            return sliced_arrays[0:hiidx+1]
    
    else:
        if len(sliced_arrays.shape) != 1:
            return sliced_arrays[:, lowidx:hiidx+1]
        else:
            return sliced_arrays[lowidx:hiidx+1]


def simulate_and_compare(varparams, varparamkeys, system, data_arrays, to, 
                         light, powers=None, irf_fwhm=40*ps, N_coarse=500, 
                         roll_value=0, comparison='linear', absolute=True, 
                         limits = None, norm=True, roll_criterion='max', 
                         maxavgnum=10, condensed_output=True):
    """
    Returns an array or list of arrays of differences between a set of
    simulated data and a set of experimental data, after aligning them. 
    The parameters and output of this functions are eventually to be passed to
    an optimization function such as ``scipy.optimize.least_sqares``.
    
    Parameters
    ----------
    varparams : list
        Rate equation parameters or other simulation parameters that determine
        the output of `system`. It does not need to be the comprehensive list 
        of parameters that define the system, simply the parameters with 
        respect to which the system output is to be optimized.
        varparams must have been obtained as list(dictionary.keys()) of the
        same dictionary that varparamkeys are extracted.
    varparamkeys : dictionary keys
        Names of varied parameters.
    system : object
        A system object to which `params` are passed, which contains a function
        that can output an array of PL signals. The system object needs to have
        been created previously in the code with dummy or guess parameters. It
        is then redefined with different parameters every time the 
        ``simulate_and_compare`` function is called.
    data_list : array or list of 1D arrays
        Array or list of 1D arrays containing experimental data to be compared 
        to the simulation. The output of ``system.PLsig`` must have the same 
        shape as `data_list`.
    to : dictionary
        Dictionary with time parameters. 
    light : object
        Excitation object that determines simulation
    powers : dictionary
        List of pulse powers at which experiment is conducted
    irf_fwhm : float, optional
        Width of Instrument Response Function used for convolution. See
        ``convolve_irf`` function. Default is 40 ps.
    N_coarse : integer
        N_coarse parameter of refined_simulation function.
    roll_value : float, optional
        Value on ``to['array']`` to which maxima of aligned arrays will be 
        shifted to.
    comparison : 'linear' or 'log'
        see ``elementwise_diff`` function. Default is linear.
    absolute : boolean, optional
        see ``elementwise_diff`` function. Default is False.
    limits : list of 2 values, optional
        lower and upper limit on time axis for which fitting is desired.
        Default is None (whole time axis fitted).
    norm : Boolean
        Determines whether each data/simulation transient should be normalized 
        with respect to itself. If false, all simulation and data traces are
        normalized by the simulation and data trace corresponding 
        to the highest power, respectively.
    roll_criterion : string, 'steep' or 'max'
        Determines whether the alignment of the two arrays happens by their 
        maximum or their steepest point. Default is 'max'
    maxavgnum : integer
        If `roll_criterion` is `"max"`, maxavgnum determines the `avgnum` 
        keyword in ``KinetiKit.kit.align_by_max().``
        
    Returns
    -------
    diffs : array or list of 1D arrays
        Array(s) of the differences or relative differences between the passed
        experimental data and simulation array(s).
    al_data_list : array or list of 1D arrays
        Aligned data arrays.
    al_sim_list : array or list of 1D arrays
        Aligned simulation arrays.
    """
      
    param_dict = kin_kit.dict_from_list(varparams, varparamkeys)
    system.update(**param_dict)
    if light is not None:
        pulse = light.pulse
    dtime = to['array'][::to['subsample']]
    
    global counter
    
    if counter%settings['display_counter_every'] == 0:
        if settings['display_counter']:
            print(counter)
    counter+=1
    
    if system.populations is None:
        pl, converged = sim.lib.simulate_func(system, dtime)
        sim_arrays = sim.lib.convolve_irf(pl, dtime, 
                                          fwhm=irf_fwhm)
    
    else:
        if powers is None:
    
            transient, converged = sim.lib.refined_simulation(system, to, light,
                                                          N_coarse=N_coarse)
            pl = system.PLsig(transient)
            sim_arrays = sim.lib.convolve_irf(pl, dtime, 
                                          fwhm=irf_fwhm)
            #sim_arrays /= max(sim_arrays)
            #data_arrays /= max(data_arrays)
            
        else:
            for i, power in enumerate(powers):
                pulse['power'] = power
                light = sim.lib.Excitation(pulse=pulse)
                transient, converged = sim.lib.refined_simulation(system, to, light,
                                                          N_coarse=N_coarse)
                pl_at_this_power = system.PLsig(transient)
                
                if i == 0:
                    pl = pl_at_this_power
                else:
                    pl = np.vstack((pl, pl_at_this_power))
                
            sim_arrays = sim.lib.convolve_irf(pl, dtime, 
                                          fwhm=irf_fwhm)    
    
    
    if roll_criterion == 'max':
        al_data_arrays = kin_kit.align_by_max(data_arrays, 
                                           dtime, 
                                           avgnum = maxavgnum,
                                           value = roll_value)
        
        al_sim_arrays = kin_kit.align_by_max(sim_arrays, 
                                           dtime,
                                           avgnum = maxavgnum,
                                           value = roll_value)
    
    elif roll_criterion == 'steep':
        al_data_arrays = kin_kit.align_by_steep(data_arrays,
                                                 dtime,
                                                 avgnum = maxavgnum,
                                                 value=roll_value)
        
        al_sim_arrays = kin_kit.align_by_steep(sim_arrays,
                                                dtime,
                                                avgnum = maxavgnum,
                                                value=roll_value)
        
    else: 
        print('Roll_criterion must be max or steep.')
    
    sim_arrays = kin_kit.make_2d(sim_arrays)
    data_arrays = kin_kit.make_2d(data_arrays)
    
    # divide all traces by maximum value of highest-power trace 
    sim_arrays /= max(sim_arrays[np.argmax(powers)]) 
    data_arrays /= max(data_arrays[np.argmax(powers)])  
    
    if limits is not None:    
        al_data_arrays = slice_by_time(al_data_arrays, dtime, limits[0], limits[1])
        al_sim_arrays = slice_by_time(al_sim_arrays, dtime, limits[0], limits[1])
    
    diffs = elementwise_diff(al_data_arrays, al_sim_arrays, 
                                 norm = norm, comparison=comparison,
                                 absolute = absolute)
        
    #print("diff : %0.3e"%(np.sum(diffs**2)/(to['N']/to['subsample'])))
    if condensed_output:
        return np.sum(diffs**2)
    else:
        return diffs.flatten()

def sac_args(varparamkeys, system, data_arrays, to, 
                         light, powers=None, irf_fwhm=40*ps, N_coarse=500, roll_value=0, 
                         comparison='linear', absolute=True, limits = None,
                         norm=True, roll_criterion='max', maxavgnum=10,
                         condensed_output=True):
    """
    Returns a list of all but the first argument of ``simulate_and_compare()``.
    Used inside a minimization function like ``differential_evolution``, which
    does not accept keyword arguments, to selectively change some of the 
    arguments.
    """
    
    return varparamkeys, system, data_arrays, to, light, powers,  irf_fwhm, \
N_coarse, roll_value, comparison, absolute, limits, norm, roll_criterion, \
maxavgnum, condensed_output

def fit_leastsq(function, p0, args):
    # original idea by https://stackoverflow.com/a/21844726
    errfunc = function(p0, *args)
    
    pfit, pcov, infodict, errmsg, success = \
        sp.optimize.leastsq(function, p0, args=args, \
                          full_output=1)

    if (len(errfunc) > len(p0)) and pcov is not None:
        s_sq = (function(pfit, *args)**2).sum()/(len(errfunc)-len(p0))
        pcov = pcov * s_sq
    else:
        pcov = np.inf

    error = [] 
    for i in range(len(pfit)):
        try:
          error.append(np.absolute(pcov[i][i])**0.5)
        except:
          error.append( 0.00 )
    pfit_leastsq = pfit
    perr_leastsq = np.array(error) 
    return pfit_leastsq, perr_leastsq, infodict


import numpy as np
import warnings

from KinetiKit.sim.systems import FunctionModel

warnings.filterwarnings("ignore", category=RuntimeWarning)

class Biexp(FunctionModel):
    """
    Biexponential fitting. To be used with sim.lib.simulate_func 
    instead of sim.lib.refined_simulation. 
    """
    
    class_name = 'Biexp'
    default = {
    'A1': 1,    # Amplitude of first lifetime
    'tau1': 1,  # First lifetime
    'tau2': 1,  # Second lifetime
    'offset':0, # y-offset
    }
        
    def PLsig(self, t):
        return self.A1*np.exp(-t/self.tau1) + (1-self.A1)*np.exp(-t/self.tau2) \
            + self.offset


"""

"""
import matplotlib
import matplotlib.pyplot as plt
# from ctypes import windll

# sdpi = windll.user32.GetDpiForSystem()

plt.style.use('bmh')
matplotlib.rcParams['axes.prop_cycle'] = matplotlib.cycler('color', ['E24A33', '348ABD', '988ED5', '8EBA42', 'FBC15E', '777777', 'FFB5B8'])
                   # E24A33 : red
                   # 348ABD : blue
                   # 988ED5 : purple
                   # 8EBA42 : green
                   # FBC15E : yellow
                   # 777777 : gray
                   # FFB5B8 : pink

# matplotlib.rcParams['figure.dpi'] = sdpi
matplotlib.rcParams['savefig.dpi'] = 800
matplotlib.rcParams['font.size'] = 10
matplotlib.rcParams['markers.fillstyle'] = 'none'
matplotlib.rcParams["errorbar.capsize"] = 3

colors = ['#E24A33', '#348ABD', '#988ED5', '#8EBA42', '#FBC15E', '#777777', '#FFB5B8']


import numpy as np
import warnings
from KinetiKit.sim.systems import RateModel
warnings.filterwarnings("ignore", category=RuntimeWarning)

class MonoHTrap(RateModel):
    """
    Single (monolithic) system, excitons, excitons trapped as holes.
    """
    
    class_name = 'MonoHTrap'
    default = {
        'k_ann': 1,
        'k_dis': 1,
        'k_rec': 1,
        'k_trx': 1,
        'k_eth': 1,
        'N_trp': 1,
        'cs': 1,
        }  
    
    populations = ['x', 'e', 'h', 'th']
   
    def rate(self, N, photons):
        nx, ne, nh, nth = N # free excitons, trapped excitons
        
        nx_rate = photons*self.cs - self.k_ann * nx - self.k_dis * nx \
        - self.k_trx * (self.N_trp - nth) * nx + self.k_rec * ne * nh
        
        ne_rate = self.k_dis * nx - self.k_rec * ne * nh \
        + self.k_trx * (self.N_trp - nth) * nx - self.k_eth * nth * ne 
        
        nh_rate = self.k_dis * nx - self.k_rec * ne * nh
        
        nth_rate = self.k_trx * (self.N_trp - nth) * nx - self.k_eth * nth * ne
        
        return np.array([nx_rate,
                         ne_rate,
                         nh_rate,
                         nth_rate])
        
    def PLsig(self, N):
        nx = N[self.populations.index('x')] 
        
        return self.k_ann * nx 



import numpy as np
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
from KinetiKit.sim.systems import RateModel

class Mono(RateModel):
    """
    Single (monolithic) system, excitons, free electrons & holes, no trapping.
    Electrons and holes recombine to produce light.
    """
    
    class_name = 'Mono'
    default = {'k_ann': 1,
               'k_dis': 1,
               'k_rec': 1,
               'cs': 1,
                    }
    populations = ['x', 'e', 'h']
    
    def rate(self, N, photons):
        nx, ne, nh = N
        nx_rate = photons*self.cs - (self.k_ann + self.k_dis) * nx
        ne_rate = self.k_dis*nx - self.k_rec*ne*nh
        nh_rate = ne_rate

        return np.array([nx_rate,
                         ne_rate,
                         nh_rate])

    def PLsig(self, N):
        nx = N[self.populations.index('x')] 
        ne = N[self.populations.index('e')]
        nh = N[self.populations.index('h')]

        return self.k_ann * nx + self.k_rec * ne * nh


class MonoRecX(RateModel):
    """
    Same as Mono, but electrons and holes can recombine into excitons, instead 
    of just producing light.
    """
    
    class_name = 'MonoRecX'
    default = {
        'k_ann': 1,
        'k_dis': 1,
        'k_rec': 1,
        'cs': 1,
        } 
    
    populations = ['x', 'e', 'h']
    
    def rate(self, N, photons):
        nx, ne, nh = N
        nx_rate = photons*self.cs - (self.k_ann + self.k_dis) * nx + self.k_rec*ne*nh
        ne_rate = self.k_dis*nx - self.k_rec*ne*nh
        nh_rate = ne_rate

        return np.array([nx_rate,
                         ne_rate,
                         nh_rate])

    def PLsig(self, N):
        nx = N[self.populations.index('x')] 

        return self.k_ann * nx
    
class MonoFracFree(RateModel):
    """
    Same as MonoRecX, but now excitation also produces a fraction of free
	carriers in addition to excitons. Electrons and holes recombine into excitons.
    """
    
    class_name = 'MonoFracFree'
    default = {'k_ann': 1,
               'k_dis': 1,
               'k_rec': 1,
               'r_free': 1,
               'cs': 1,
                    }
    populations = ['x', 'e', 'h']
    
    def rate(self, N, photons):
        nx, ne, nh = N
        nx_rate = photons*self.cs*(1-self.r_free) - (self.k_ann + self.k_dis)*nx + self.k_rec*ne*nh
        ne_rate = photons*self.cs*self.r_free + self.k_dis*nx - self.k_rec*ne*nh
        nh_rate = ne_rate

        return np.array([nx_rate,
                         ne_rate,
                         nh_rate])

    def PLsig(self, N):
        nx = N[self.populations.index('x')] 

        return self.k_ann * nx



    
    
